import socket
from typing import List, Optional

from .buffer import Buffer
from .connection import Connection
from .exceptions import PyQ3SLError
from .server import Server


class PrincipalServer:
    address: str
    port: int
    connection: Connection

    def __init__(
            self,
            address: str,
            port: int,
            network_protocol: socket.SocketKind = socket.SOCK_DGRAM,
            timeout: float = 1.0
    ):
        self.address = address
        self.port = port
        self.connection = Connection(self.address, self.port, network_protocol, timeout)

    def get_servers(
            self,
            query_protocol: int,
            game_name: str = '',
            keywords: str = 'full empty',
            server_entry_prefix: Optional[bytes] = None
    ) -> List[Server]:
        buffer = Buffer(b'\xff' * 4)
        buffer.write_string('getservers ')
        # Add game name if set
        if game_name != '':
            buffer.write_string(f'{game_name} ')
        # Add query protocol and keywords
        buffer.write_string(f'{query_protocol} {keywords}')

        self.connection.write(buffer.get_buffer())
        result = self.connection.read()

        # Server entries are separated by slashes
        return self.parse_response(result, b'\\', server_entry_prefix)

    @staticmethod
    def parse_response(buffer: Buffer, sep: Optional[bytes] = None, prefix: Optional[bytes] = None) -> list:
        if not PrincipalServer.buffer_has_packet_header(buffer, sep):
            raise PyQ3SLError('Principal returned invalid data')

        sep_len = len(sep) if sep is not None else 0
        """
        Some 3rd party implementations of the protocol also prefix every server entry with the same
        byte sequence (CoD4 X for example) => ignore prefix while parsing
        """
        prefix_len = len(prefix) if prefix is not None else 0

        # Servers are represented as six byte sequences, plus the length of any separator and prefix
        servers = []
        while buffer.has(6 + sep_len + prefix_len):
            # Skip backspace delimiter and prefix
            buffer.skip(sep_len + prefix_len)
            ip, port = buffer.read_ip(), buffer.read_ushort()

            """
            Some principals return complete responses in each packet
            => check for and skip b'\\EOT' and additional headers
            """
            if buffer.peek(4) == b'\\EOT':
                buffer.skip(4)
            PrincipalServer.buffer_has_packet_header(buffer, sep)

            # Init and append valid server
            if ip != '0.0.0.0' and port != 0:
                servers.append(Server(ip, port))

        return servers

    @staticmethod
    def buffer_has_packet_header(buffer: Buffer, sep: Optional[bytes] = None) -> bool:
        if buffer.peek(22) != (b'\xff' * 4 + b'getserversResponse'):
            return False

        buffer.skip(22)

        """
        Some principals send a few extra bytes before the first server delimiter, Activision for example sends
        b'\n\x00' => read until we see the first delimiter
        """
        while sep is not None and buffer.peek(1) != sep:
            buffer.skip(1)

        return True



import re
import struct
from typing import Union, List

from .exceptions import PyQ3SLError

COLOR_REGEX = re.compile(r'\^(X.{6}|.)')


class Buffer:
    data: bytes
    length: int
    index: int

    def __init__(self, data: bytes = b''):
        self.data = data
        self.length = len(data)
        self.index = 0

    def get_buffer(self) -> bytes:
        return self.data[self.index:]

    def read(self, length: int = 1) -> bytes:
        if self.index + length > self.length:
            raise PyQ3SLError('Attempt to read beyond buffer length')

        data = self.data[self.index:self.index + length]
        self.index += length

        return data

    def peek(self, length: int = 1) -> bytes:
        return self.data[self.index:self.index + length]

    def skip(self, length: int = 1) -> None:
        self.index += length

    def has(self, length: int = 1) -> bool:
        return len(self) >= length

    def __len__(self):
        return max(0, self.length - self.index)

    def read_string(
            self,
            sep: Union[bytes, List[bytes]],
            consume_sep: bool = False,
            encoding: str = 'latin1',
            strip_colors: bool = True
    ) -> str:
        b = self.get_buffer()
        """
        ioquake3 server may contain an "fs_manifest", which contains "\n " as a delimiter. Since "\n" would usually
        terminate the sever info like, this breaks the format. So, "greedily" try to use the first given delimiter,
        only switch alternate ones if previous cannot be found at all (will read past the "\n " in "fs_manifest" since 
        it is not the last value.
        """
        sep_list = [sep] if type(sep) == bytes else sep
        index = next((i for sep in sep_list if (i := b.find(sep)) != -1), -1)
        if index == -1:
            raise PyQ3SLError('Expected string delimiters were not found')
        v = self.read(index)

        if consume_sep:
            self.skip(1)

        raw = v.decode(encoding, errors='replace')
        if strip_colors:
            return COLOR_REGEX.sub('', raw)

        return raw

    def read_ushort(self) -> int:
        v, *_ = struct.unpack('>H', self.read(2))
        return v

    def read_ip(self) -> str:
        v = self.read(4)
        return "%d.%d.%d.%d" % struct.unpack(">BBBB", v)

    def write(self, v: bytes) -> None:
        self.data += v
        self.length += len(v)

    def write_string(self, v: str, encoding: str = 'latin1') -> None:
        self.write(v.encode(encoding))

    def write_ushort(self, v: int) -> None:
        self.write(struct.pack('>H', v))



import socket
from typing import Any

from .buffer import Buffer
from .connection import Connection
from .exceptions import PyQ3SLError


class Server:
    ip: str
    port: int

    def __init__(self, ip: str, port: int):
        self.ip = ip
        self.port = port


    def __repr__(self):
        return f'{self.ip}:{self.port}'

    def __iter__(self):
        yield 'ip', self.ip
        yield 'port', self.port

    def __eq__(self, other: Any):
        return isinstance(other, type(self)) and \
            other.ip == self.ip and \
            other.port == self.port

    def get_status(self, strip_colors: bool = True, timeout: float = 1.0):
        connection = Connection(self.ip, self.port, socket.SOCK_DGRAM, timeout)

        packet = self.build_query_packet()

        connection.write(packet)
        result = connection.read()
        return self.parse_response(result, strip_colors)

    def parse_response(self, buffer: Buffer, strip_colors: bool) -> dict:
        """
        Response should consist of at least three lines:
        1: header indicating response type
        2: list of server variables, delimited by \
        3+: lines containing player info (final player line being empty)
        """
        # Make sure header indicates status response as type
        if not self.has_valid_response_header(buffer):
            raise PyQ3SLError('Server returned invalid packet header')

        # Make sure body starts with "\" indicating the first variable and contains an even number of keys and values
        if not self.has_valid_response_body(buffer):
            raise PyQ3SLError('Server returned invalid packet body')

        # Parse variable keys and values
        i = 0
        keys = []
        values = []
        while buffer.peek(1) == b'\\':
            """
            Skip the \\ indicating the start of the key/value and use
            a) all bytes until the next separator (anything but the last value
            or
            b) all bytes until the next linebreak (last value)
            as the key/value
            """
            buffer.skip(1)
            element = buffer.read_string([b'\\', b'\n'], strip_colors=strip_colors)
            if i % 2 == 0:
                keys.append(element)
            else:
                values.append(element)

            i += 1

        players = []
        # A player with 0 frags, 0 ping and an empty name would take up 7 bytes plus the line break
        while buffer.peek(1) == b'\n' and buffer.has(8):
            player = self.parse_player(buffer, strip_colors)
            players.append(player)

        return {
            'ip': self.ip,
            'port': self.port,
            **dict(zip(keys, values)),
            'players': players
        }

    @staticmethod
    def build_query_packet() -> bytes:
        return b'\xff\xff\xff\xffgetstatus\x00'

    @staticmethod
    def has_valid_response_header(buffer: Buffer) -> bool:
        return buffer.has(19) and buffer.read(19) == b'\xff\xff\xff\xffstatusResponse\n'

    @staticmethod
    def has_valid_response_body(buffer: Buffer) -> bool:
        return buffer.peek(1) == b'\\' and buffer.get_buffer().count(b'\\') % 2 == 0

    @staticmethod
    def parse_player(buffer: Buffer, strip_colors: bool) -> dict:
        frags = int(buffer.read_string(b' ', consume_sep=True, strip_colors=strip_colors))
        ping = int(buffer.read_string(b' ', consume_sep=True, strip_colors=strip_colors))
        buffer.skip(1)
        name = buffer.read_string(b'"', consume_sep=True, strip_colors=strip_colors)

        return {
            'frags': frags,
            'ping': ping,
            'name': name,
        }


class MedalOfHonorServer(Server):
    """
    Medal of Honor uses a custom Quake3 protocol variant with some key differences, making it incompatible with the
    "vanilla" protocol. Since all the Medal of Honor games use GameSpy to list servers, Medal of Honor servers can only
    be created directly.
    """
    def __init__(self, ip: str, port: int):
        super().__init__(ip, port)

    @staticmethod
    def build_query_packet() -> bytes:
        # Medal of Honor uses a slightly different query packet
        return b'\xff\xff\xff\xff\x02getstatus xxx\x00'

    @staticmethod
    def has_valid_response_header(buffer: Buffer) -> bool:
        # Medal of Honor responses contain an extra byte (b'\x01')
        return buffer.has(20) and buffer.read(20) == b'\xff\xff\xff\xff\x01statusResponse\n'

    @staticmethod
    def parse_player(buffer: Buffer, strip_colors: bool) -> dict:
        # Medal of Honor only sends a player's ping and name (seems like colors are not supported)
        ping = int(buffer.read_string(b' ', consume_sep=True, strip_colors=strip_colors))
        buffer.skip(1)
        name = buffer.read_string(b'"', consume_sep=True, strip_colors=strip_colors)

        return {
            'ping': ping,
            'name': name,
        }



import socket

from .buffer import Buffer
from .exceptions import PyQ3SLError, PyQ3SLTimeoutError


class Connection:
    address: str
    port: int
    protocol: socket.SocketKind
    sock: socket.socket
    timeout: float
    is_connected: bool

    def __init__(self, address: str, port: int, protocol: socket.SocketKind, timeout: float):
        self.address = address
        self.port = port
        self.protocol = protocol
        self.timeout = timeout

        self.is_connected = False

    def connect(self) -> None:
        if self.is_connected:
            return

        self.sock = socket.socket(socket.AF_INET, self.protocol)
        self.sock.settimeout(self.timeout)

        try:
            self.sock.connect((self.address, self.port))
            self.is_connected = True
        except socket.timeout:
            self.is_connected = False
            raise PyQ3SLTimeoutError(f'Connection attempt to {self.address}:{self.port} timed out')
        except socket.error as e:
            self.is_connected = False
            raise PyQ3SLError(f'Failed to connect to {self.address}:{self.port} ({e})')

    def write(self, data: bytes) -> None:
        if not self.is_connected:
            self.connect()

        try:
            self.sock.sendall(data)
        except socket.error:
            raise PyQ3SLError('Failed to send data to server')

    def read(self) -> Buffer:
        if not self.is_connected:
            self.connect()

        data = b''
        last_packet_length = 0
        receive_next = True

        while receive_next:
            try:
                # Packet size differs from server to server => just read up to max possible UDP size
                iteration_data = self.sock.recv(65507)
            except socket.timeout:
                # Raise exception if no data was retrieved at all, else break loop
                if data == b'':
                    raise PyQ3SLTimeoutError('Timed out while receiving server data')
                else:
                    break
            except socket.error:
                raise PyQ3SLError('Failed to receive data from server')

            data += iteration_data

            """
            Continue to try reading from socket until
            a) packets get shorter (UDP socket) or
            b) peer indicates EOF/stops returning new data (TCP socket)
            """
            buffer_end = iteration_data[-10:]
            receive_next = (self.protocol == socket.SOCK_DGRAM and len(iteration_data) >= last_packet_length) or \
                           (self.protocol == socket.SOCK_STREAM and b'EOF' not in buffer_end and len(iteration_data) != 0)
            last_packet_length = len(iteration_data)

        return Buffer(data)

    def __del__(self):
        self.close()

    def close(self) -> bool:
        if hasattr(self, 'sock') and isinstance(self.sock, socket.socket):
            if self.is_connected:
                self.sock.shutdown(socket.SHUT_RDWR)
            self.sock.close()
            self.is_connected = False
            return True

        return False



from .connection import Connection
from .exceptions import PyQ3SLError, PyQ3SLTimeoutError
from .principalserver import PrincipalServer
from .server import Server, MedalOfHonorServer

"""
pyq3serverlist.

Simple Python library for querying Quake 3 based principal servers and their game servers.
"""

__version__ = '0.3.1'
__author__ = 'cetteup'
__credits__ = 'https://github.com/jacklul'
__all__ = ['Connection', 'PrincipalServer', 'Server', 'MedalOfHonorServer',  'PyQ3SLError', 'PyQ3SLTimeoutError']
